@@ -22,6 +22,8 @@ var default_settings = {
 	staged: false,
 	unstaged: false,
 	show_parents: false,
+	use_gravatar: true,
+	default_collapse_all: true,
 	strings: {
 		stage: 'stage',
 		unstage: 'unstage',
@@ -29,7 +31,9 @@ var default_settings = {
 		notes: 'Notes:',
 		parents: 'Parents:',
 		diff_against: 'Diff against:',
-		committed_by: 'Committed by:'
+		committed_by: 'Committed by:',
+		expand_all: 'Expand all',
+		collapse_all: 'Collapse all'
 	},
 };
 
@@ -79,13 +83,21 @@ function get_commit_elements(content)
 	return commit_elements;
 }
 
+var avatarLoader;
+
 function write_avatar(avatar, commit)
 {
 	var h = commit.author.email_md5;
+	var avc;
 
-	if (h in avatar_cache)
+	if (avatarLoader)
 	{
-		var avc = avatar_cache[h];
+		avatarLoader.cancelled = true;
+	}
+
+	if (h in avatar_cache && settings.use_gravatar)
+	{
+		avc = avatar_cache[h];
 
 		if (avc != null && avatar.attr('src') != avc)
 		{
@@ -95,23 +107,42 @@ function write_avatar(avatar, commit)
 		return;
 	}
 
-	var loader = $('<img/>');
-	var gravatar = 'http://www.gravatar.com/avatar/' + h + '?d=404&s=50';
+	var avc = 'gitg-diff:/icon/avatar-default-symbolic?size=60';
+	avatar.attr('src', avc);
 
-	loader.on('load', function () {
-		avatar_cache[h] = gravatar;
+	if (settings.use_gravatar)
+	{
+		avatarLoader = {
+			image: $('<img/>'),
+			cancelled: false
+		};
 
-		avatar.attr('src', gravatar);
-	});
+		var gravatar = 'http://www.gravatar.com/avatar/' + h + '?d=404&s=60';
 
-	loader.on('error', function () {
-		var avc = 'gitg-diff:/icon/avatar-default-symbolic?size=50';
+		avatarLoader.image.on('load', function () {
+			if (this.cancelled)
+			{
+				return;
+			}
 
-		avatar_cache[h] = avc;
-		avatar.attr('src', avc);
-	});
+			avatar_cache[h] = gravatar;
+			avatar.attr('src', gravatar);
 
-	loader.attr('src', gravatar);
+			avatarLoader = null;
+		}.bind(avatarLoader));
+
+		avatarLoader.image.on('error', function () {
+			if (this.cancelled)
+			{
+				return;
+			}
+
+			avatar_cache[h] = avc;
+			avatarLoader = null;
+		}.bind(avatarLoader));
+
+		avatarLoader.image.attr('src', gravatar);
+	}
 }
 
 function open_url(target)
@@ -187,7 +218,7 @@ function write_commit(content, commit)
 	}
 	else
 	{
-		elems['committed-by'].html('');
+		elems['committed-by'].html('&nbsp;<br>&nbsp;');
 	}
 
 	// Date
@@ -304,6 +335,62 @@ function expand_collapse()
 	}
 
 	expander.closest('tbody').toggleClass("collapsed");
+	collapsed_changed();
+}
+
+function collapsed_changed()
+{
+	var all = $("#diff_content div.file tbody");
+	var prevCollapsed = false;
+	var allUncollapsed = true;
+
+	all.each(function(i, tbody) {
+		tbody = $(tbody);
+
+		var nextCollapsed = false;
+
+		var beforeCollapsed = true;
+		var afterCollapsed = false;
+
+		if (tbody.hasClass("collapsed"))
+		{
+			if (i < all.length - 1)
+			{
+				nextCollapsed = $(all[i + 1]).hasClass("collapsed");
+			}
+
+			afterCollapsed = prevCollapsed;
+			beforeCollapsed = nextCollapsed;
+
+			prevCollapsed = true;
+			allUncollapsed = false;
+		}
+		else
+		{
+			prevCollapsed = false;
+		}
+
+		if (!beforeCollapsed)
+		{
+			tbody.addClass("before-uncollapsed");
+		}
+		else
+		{
+			tbody.removeClass("before-uncollapsed");
+		}
+
+		if (!afterCollapsed)
+		{
+			tbody.addClass("after-uncollapsed");
+		}
+		else
+		{
+			tbody.removeClass("after-uncollapsed");
+		}
+	});
+
+	var expanderAll = $("#diff .expander-all");
+	expanderAll.text(allUncollapsed ? settings.strings.collapse_all : settings.strings.expand_all)
 }
 
 function next_element(elem)
@@ -503,6 +590,40 @@ function update_tab_width(width)
 	tab_width_rule.style.tabSize = width;
 }
 
+function expand_collapse_all()
+{
+	var collapse = ($(this).text() === settings.strings.collapse_all);
+
+	var allbodies = document.querySelectorAll("#diff_content div.file:not(.background) table.file tbody");
+
+	for (var i = 0; i < allbodies.length; i++)
+	{
+		var tbody = $(allbodies[i]);
+		var isCollapsed = tbody.hasClass("collapsed");
+
+		if (isCollapsed !== collapse)
+		{
+			expand_collapse.call(tbody.find(".expander"));
+		}
+	}
+
+	$(this).text(collapse ? settings.strings.expand_all : settings.strings.collapse_all);
+}
+
+function default_collapse_all()
+{
+	var expanders = document.querySelectorAll("#diff_content div.file:not(.background) .expander");
+
+	// Collapse by default if more than one file
+	if (expanders && expanders.length > 1)
+	{
+		for (var i = 0; i < expanders.length; i++)
+		{
+			expand_collapse.call(expanders[i]);
+		}
+	}
+}
+
 function update_diff(id, lsettings)
 {
 	if (html_builder_worker)
@@ -580,7 +701,16 @@ function update_diff(id, lsettings)
 			content.html(event.data.diff_html);
 			update_has_selection();
 
-			$(".expander").click(expand_collapse);
+			if (settings.default_collapse_all)
+			{
+				default_collapse_all();
+			}
+
+			collapsed_changed();
+
+			$("#diff_content div.file tr.file_header td").click(function() {
+				expand_collapse.call($(this).find(".expander"));
+			});
 
 			if (settings.staged || settings.unstaged)
 			{
@@ -622,6 +752,9 @@ function update_diff(id, lsettings)
 }
 
 addEventListener('DOMContentLoaded', function () {
+	var expanderAll = $("#diff .expander-all");
+	expanderAll.click(expand_collapse_all);
+
 	xhr_get('internal', {action: 'loaded'});
 }, false);
 
